
#ifndef __KERNEL__
 #define __KERNEL__
#endif

#ifndef MODULE
 #define MODULE
#endif

#include <linux/config.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/fcntl.h>
#include <linux/ioctl.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <asm/system.h>
#include <asm/segment.h>
#include <asm/io.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/pci.h>

#include "dpci8255.h"
#include "sysdep.h"   

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
#include <asm/uaccess.h>
#endif

/* I defined some macro for device_open function */
#define TYPE(dev) (MINOR(dev) >> 4) 
#define NUM(dev) (MINOR(dev) & 0x0f)

#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a)*65536+(b)*256+(c))
#endif

#ifndef DECISION_VENDOR_ID
#define DECISION_VENDER_ID 0x6666
#endif

#ifndef DECISION_DEVICE_ID
#define DECISION_DEVICE_ID 0x0101
#endif


#define AINMASK  0x10
#define BINMASK  0x02
#define CINMASK  0x09



static int cards;
static char rev[2];
static int direction[MAXCARD][6];
static int counterbcd[MAXCARD][3];
static int countermode[MAXCARD][3];
int Base_Addr[MAXCARD];

/*---------------------------------------------------------------------*/
int dpci8255_open(struct inode *inoder,struct file *flip)
{
/* int type= TYPE(inoder->i_rdev);*/

 int num=NUM(inoder->i_rdev);
 if(num>=cards) return -ENODEV;

 MOD_INC_USE_COUNT;
 return 0;
}
/*---------------------------------------------------------------------*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static int dpci8255_release(struct inode *inoder,struct file *flip)
#else
static void dpci8255_release(struct inode *inoder,struct file *flip)
#endif
{
  MOD_DEC_USE_COUNT;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
  return 0;
#endif    
}
/*---------------------------------------------------------------------*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static ssize_t dpci8255_read(struct file *flip,char *buf,size_t count,loff_t *ptr)
#else
static int dpci8255_read(struct inode *inode,struct file *flip, char *buf,count_t count)
#endif
{
 int countt=0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
  int num=MINOR(flip->f_dentry->d_inode->i_rdev);
  if(ptr!=&flip->f_pos)  /* can't seek on this device */
   return -ESPIPE;
#else
  int num=NUM(inode->i_rdev);
#endif 
 
 if(num>=cards)
   return -EINVAL;
 buf[0]=readb(Base_Addr[num]);
 countt=1;
           
 return countt;
}
/*---------------------------------------------------------------------*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static ssize_t dpci8255_write(struct file *flip,const char *buf,size_t count,loff_t *ptr)
#else
static int dpci8255_write(struct inde *inode,struct file *flip,char *buf,count_t count)
#endif
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)	
 if(ptr!=&flip->f_pos)  /* can't seek on this device */
  return -ESPIPE;
#endif
 
 return 0;       /* do nothing */
}
/*---------------------------------------------------------------------*/
int Myread(int num,char *buf,int count)
{
 int countt;
 switch(buf[0])
  {
   case 0:                /* the first 8255  */
           buf[2]=inb(Base_Addr[num]+(int)buf[1]);
           countt=3;
           break;
   case 3:                /* control word */
           buf[2]=inb(Base_Addr[num]+(int)buf[1]*4+3);
           countt=3;
           break;
   case 1:                /* the second 8255 */    
           buf[2]=inb(Base_Addr[num]+(int)buf[1]+4);
           countt=3;
           break;
   case 2:                /* 8253 */
           buf[2]=inb(Base_Addr[num]+(int)buf[1]+8);
           countt=3;
           break;
    default:
           countt=0;
           break;
  }
 return countt;
}
/*---------------------------------------------------------------------*/
void Mywrite(int num,char *buf,int count)
{
 switch(buf[0])
  {
   case 0:                /* the first 8255 */
           outb(buf[2],Base_Addr[num]+(int)buf[1]);   
           break;
   case 3:                /* control word */ 
           outb(buf[2],Base_Addr[num]+(int)buf[1]*4+3);
           break;
   case 1:                /* the second 8255 */
           outb(buf[2],Base_Addr[num]+(int)buf[1]+4);
           break;
   case 2:                /* 8253 */
           outb(buf[2],Base_Addr[num]+(int)buf[1]+8);
           break;
   default:
           break;
  }
}
/*---------------------------------------------------------------------*/
int dpci8255_ioctl(
    struct inode *inode,
    struct file *file,
    unsigned int ioctl_num,/* The number of the ioctl */
    unsigned long ioctl_param) /* The parameter to it */
{
  char param[4];
  char *temp=(char *)ioctl_param;
  char ch1=0;
  int data;
  int num=NUM(inode->i_rdev);
  switch (ioctl_num) 
   {

    case DPCI8255_SET_REGISTER:
      ch1=(char)((int)temp);
      if(ch1<3)
       {
        rev[0]=0;
        rev[1]=ch1;
        return 0;
       }
      if(ch1<6)
       {
        rev[0]=1;
        rev[1]=(ch1-3);
        return 0;
       }
      if(ch1<9)
       {
        rev[0]=2;
        rev[1]=(ch1-6);
        return 0;
       } 
      if((ch1>9)&&(ch1<13))
       {
        rev[0]=3;
        rev[1]=(ch1-10);
       } 
      break;

    case DPCI8255_WRITE_PORT:
      param[0]=rev[0];
      param[1]=rev[1];
      param[2]=(char)((int)temp);
      Mywrite(num,param,4);
      break;

    case DPCI8255_READ_PORT:
      param[0]=rev[0];
      param[1]=rev[1];
      Myread(num,param,4);
      put_user(param[2],temp);
      break;

    case DPCI8255_READ_COUNTER:
      param[0]=2;
      param[1]=rev[1];
      Myread(num,param,4);
      put_user(param[2],temp);
      break; 
      
    case DPCI8255_WRITE_COUNTER:
      data=(int)ioctl_param;
      param[0]=2;
      param[1]=rev[1];
      param[2]=(char)data;
      param[3]=(char)(data>>8);
      Mywrite(num,param,4);
      break;

    case DPCI8255_SET_CONTROLWORD:
      param[0]=3;
      param[1]=rev[1];
      param[2]=(char)((int)temp);
      Mywrite(num,param,4);
      break;      
            
    case DPCI8255_GET_CONTROLWORD:
      param[0]=3;
      param[1]=rev[1];
      Myread(num,param,4);
      put_user(param[2],temp);
      break;
     default:
             break;
  }

  return 0;
}
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/
static struct file_operations dpci8255_fops={
read: dpci8255_read,
write: dpci8255_write,
ioctl: dpci8255_ioctl ,
open: dpci8255_open,
release: dpci8255_release,
};
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/
MODULE_AUTHOR("Delta");
MODULE_DESCRIPTION("Decision PCI 8255 Driver");
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/

static int __init dpci8255_init(void)
{
 int index=0;
 struct pci_dev *dev = NULL;
 static int dev_count=0;
 int i; 

 dev_count++;
 if(dev_count>1)
  {
   printk(KERN_ERR"This drive only can open one time.\n");
   return -ENODEV;
  }

#ifdef CONFIG_PCI
 if (!pci_present())   /* No PCI bus in this machine! */
#endif
 return -ENODEV;

 cards=0;
 while (index < MAXCARD && /* dev == NULL && */
        (dev = pci_find_device(DECISION_VENDER_ID, DECISION_DEVICE_ID,dev))) 
 {
/*
  for(i=0;i<6;i++)
   {
    printk("PCI 8255 Base[%d] : %x\n",i,(int)(dev->base_address[i] & PCI_BASE_ADDRESS_IO_MASK));
   } 
 */   
   Base_Addr[index]=pci_resource_start(dev,3) & PCI_BASE_ADDRESS_IO_MASK;   
   printk("PCI 8255:I got a base %x\n",Base_Addr[index]); 
   if(check_region(Base_Addr[index],REGIONLENGTH)>=0)
    request_region(Base_Addr[index],REGIONLENGTH,DEVICE_NAME);
   
   for(i=0;i<6;i++)
    direction[index][i]=0;  // default is output
   for(i=0;i<3;i++)
    {
     countermode[index][i]=3;
     counterbcd[index][i]=0;
    } 
/*   outb(0,Base_Addr[index]+3);
   outb(0,Base_Addr[index]+7); 
   outb(0x36,Base_Addr[index]+11);
   outb(0xff,Base_Addr[index]+8);
   outb(0xff,Base_Addr[index]+8);
*/  
   index++;
  }

  if((dev==NULL)&&(index<1))
   {
    printk("\nI can't find any Decision PCI 8255 card here!!\n");
    return -ENODEV;
   }
 printk("\nPCI 8255: Total %d cards that I catched!\n",index);
 cards=index;
 index=register_chrdev(dpci8255_major,DEVICE_NAME,&dpci8255_fops);
 if(index<0)
  {
  for(i=0;i<cards;i++)
   release_region(Base_Addr[i],REGIONLENGTH);
  return index;
  }
 rev[0]=0;
 rev[1]=0;
 printk("PCI 8255:Register Successful\n");  
 return 0;

}
/*---------------------------------------------------------------------*/
static void __exit dpci8255_exit(void)
{
  int i;
  for(i=0;i<cards;i++)
   release_region(Base_Addr[i],REGIONLENGTH); 
  unregister_chrdev(dpci8255_major,DEVICE_NAME);
  printk("PCI 8255:Unregister Successful!!\n");  
}
/*---------------------------------------------------------------------*/
module_init(dpci8255_init);
module_exit(dpci8255_exit);
